package eos.agents;

import java.util.*;

import eos.framework.Economy;
import eos.framework.Good;
import eos.framework.Market;
import eos.models.BaselineModel.*;

/**
 * SimpleFarm2 is a modification of SimpleFarm which buys food using a basic
 * profit-maximizing technique based on the last market price for food and sells
 * using a very simple price = f(food on hand) technique. Exponential smoothing
 * is used to create a base-price for selling food.
 * 
 * @author Michael Adelson
 */
public class SimpleFarm extends BaselineFarm {

    /**
     * the increments in which the farm bids on labor
     */
    private static final double DISCREET_LABOR_BID = 1.0;

    /**
     * the farm keeps bidding on labor until the expected revenue falls below
     * this amount
     */
    private static final double MIN_REVENUE = 0.1;

    /**
     * If the farm has more than this amount of food, it will lower it's asking
     * price. If it has less, it will raise it.
     */
    private static final double FOOD_THRESHOLD = 60.0;

    /**
     * Alpha used in exponential smoothing.
     */
    private static final double ALPHA = 0.1;

    /**
     * The exponentially smoothed value estimate for food.
     */
    private double value;

    /**
     * The economy's food and labor for money markets.
     */
    private Market foodMarket, laborMarket;

    /**
     * Create a new DemandCurveFarm in economy.
     */
    public SimpleFarm(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
        foodMarket = null;
        laborMarket = null;
    }

    /**
     * bid on labor and sell food.
     */
    protected void bid() {
        // update value of food
        if (foodMarket == null) { // initialize
            foodMarket = economy.getMarketFor(money.getName(), food.getName());
            laborMarket = economy
                    .getMarketFor(money.getName(), labor.getName());
            value = this.economy.random() + 1e-3;
        } else
            value = ALPHA * foodMarket.getLastMarketPrice() + (1 - ALPHA)
                    * value;

        bidOnLabor();
        bidOnFood();
    }

    /**
     * Bid on labor by creating a demand curve for it, using the estimated
     * market price of food to calculate expected revenue from the hired labor.
     */
    private void bidOnLabor() {
        double i = DISCREET_LABOR_BID;

        /**
         * Offer bids until the farm runs out of money or the expected revenue
         * gets too low. This basically creates a demand curve for labor that
         * gets its downward slope from the diminishing returns of
         * convertFoodToLabor()
         */
        while (money.getQuantity() > 0.0) {

            double foodToProduce = convertToFood(i)
                    - convertToFood(i - DISCREET_LABOR_BID);
            double expectedRevenue = foodToProduce
                    * foodMarket.getLastMarketPrice();
            if (expectedRevenue < MIN_REVENUE) {
                break;
            }
            if (expectedRevenue > money.getQuantity()) {
                expectedRevenue = money.getQuantity();
            }
            double unitPrice = expectedRevenue / DISCREET_LABOR_BID;
            // System.out.println("money: " + money.getQuantity() +
            // " labor price: " + unitPrice);
            laborMarket.addBuyOffer(this, DISCREET_LABOR_BID, unitPrice);

            i += DISCREET_LABOR_BID;

        }

    }

    /**
     * Set food prices based on quantity and historic market price.
     */
    private void bidOnFood() {

        // cannot sell food
        if (food.getQuantity() <= 0.0) {
            return;
        }

        /*
         * if lastPrice = 0, sell food at various prices, effectively "probing"
         * the Laborer's demand curves.
         */
        if (foodMarket.getLastMarketPrice() <= 0.0) {
            double maxBids = 30, Qf = food.getQuantity();
            for (double d = 0; d < maxBids; d++)
                foodMarket.addSellOffer(this, Qf / d, this.economy.random()
                        * value);
            return;
        }

        // double value = foodMarket.getLastMarketPrice();

        // else adjust price based on the farm's quantity of food
        double price = value;
        price *= FOOD_THRESHOLD / food.getQuantity();
        // System.out.println("food: " + food.getQuantity() + " price " +
        // price);
        foodMarket.addSellOffer(this, food.getQuantity(), price);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.framework.Agent#toString()
     */
    public String toString() {
        return super.toString() + "Estimated value of labor = " + this.value
                + "\n";
    }

}
