/**
 * 
 */
package eos.datatypes;

import java.util.*;

/**
 * Represents a value to be exponentially smoothed. Smoothes via the update()
 * function, setting value = (alpha * newValue) + (1 - alpha) * value.
 * 
 * @author Michael Adelson
 */
public class SmoothedValue extends Number implements Comparable<Number> {

    /**
     * Default serial version id.
     */
    private static final long serialVersionUID = 1L;

    /**
     * The current smoothed value
     */
    private double value;

    /**
     * The smoothing coefficient.
     */
    private double alpha;

    /**
     * Construct a SmoothedValue starting at startingValue and using alpha as
     * the smoothing coefficient.
     */
    public SmoothedValue(double startingValue, double alpha) {
        this.value = startingValue;
        this.alpha = alpha;
    }

    /**
     * Updates value: value = (alpha * newValue) + (1 - alpha) * value.
     */
    public final void update(double newValue) {
        if (this.value == newValue)
            return;
        this.value = (this.alpha * newValue) + (1 - this.alpha) * this.value;
    }

    /**
     * @return a new SmoothedValue object with the same alpha and value as this
     *         object.
     */
    public final SmoothedValue copy() {
        return new SmoothedValue(this.value, this.alpha);
    }

    /**
     * Changes the smoothing constant to newAlpha.
     */
    public final void setAlpha(double newAlpha) {
        this.alpha = newAlpha;
    }

    /**
     * @return the smoothing constant for this SmoothedValue.
     */
    public final double getAlpha() {
        return this.alpha;
    }

    /**
     * Changes the current value of this SmoothedValue to newValue.
     */
    public final void setValue(double newValue) {
        this.value = newValue;
    }

    /**
     * @return the value of this SmoothedValue.
     */
    public final double getValue() {
        return this.value;
    }

    /**
     * Compare this.value to that. Works for all Numbers including
     * SmoothedValues.
     */
    public final int compareTo(Number that) {
        if (that instanceof SmoothedValue)
            return Double.compare(this.value, ((SmoothedValue) that).value);
        else
            return Double.compare(this.value, that.doubleValue());
    }
 
    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Number#doubleValue()
     */
    public final double doubleValue() {
        return this.value;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Number#floatValue()
     */
    public final float floatValue() {
        return (float) this.value;
    }

    /**
     * Returns this value rounded to an integer.
     */
    public final int intValue() {
        return (int) this.longValue();
    }

    /**
     * Returns this value rounded to a long.
     */
    public final long longValue() {
        return Math.round(this.value);
    }

    /* ------------------ TEST CLIENT, DEPRECATED METHODS --------------------*/
    
    /**
     * Test client
     */
    public static void main(String[] args) {
        System.out.println("A simple test: ");
        Random r = new Random();
        double[] d = new double[10];
        for (int i = 0; i < d.length; i++)
            d[i] = r.nextDouble();
        System.out.println("Smoothing these values starting from 0.5, using alpha = 0, 0.5, and 1");
        System.out.println(Arrays.toString(d));
        for (int i = 0; i < 3; i++) {
            double[] out = new double[d.length];
            double alpha = i / 2.0;
            SmoothedValue sv = new SmoothedValue(0.5, alpha);
            for (int j = 0; j < d.length; j++) {
                sv.update(d[j]);
                out[j] = sv.getValue();
            }
            System.out.println("Alpha = " + alpha + ": " + Arrays.toString(out));
        }
        System.out.println();
        
        int base = r.nextInt(51) - 25;
        System.out
                .println("Smoothing the function " + base + " + random noise");
        System.out.println("Next Value\tSmoothed Value");

        SmoothedValue val = null;
        for (int i = 0; i < 30; i++) {
            double nextValue = base + 0.25 * r.nextDouble() * base * Math.signum(r.nextDouble() - 0.5);
            if (val == null)
                val = new SmoothedValue(nextValue, 0.1);
            else
                val.update(nextValue);
            System.out.printf("%3f\t%3f\r\n", nextValue, val.getValue());
        }
    }
}
