package eos.economies;

import java.util.*;
import eos.framework.*;
import eos.io.*;
import eos.models.SingleOwnerFirms.*;
import static eos.utilities.Errors.err;

/**
 * An economy class that implements DataOutputEconomy for output, for use in my
 * Fall 2009 independent work.
 * 
 * By Michael Adelson
 */
public class BL321Economy extends Economy implements DataOutputEconomy,
        BuildableEconomy {

    // holds all of the Agents in this Economy
    private LinkedHashSet<Agent> agents;

    // holds all of the dead Agents that have died this step
    private ArrayList<Agent> deadAgents;

    /**
     * holds all markets in a way that can be efficiently iterated and searched.
     * The key used is simply the concatenation of the currency and product
     * class names (using Class.toString()).
     */
    private LinkedHashMap<String, Market> markets;

    // holds all printers
    private List<Printer> printers;

    /**
     * Create an empty BaselineEconomy.
     */
    public BL321Economy() {
        super();
        this.agents = new LinkedHashSet<Agent>();
        this.deadAgents = new ArrayList<Agent>();
        this.markets = new LinkedHashMap<String, Market>();
        this.printers = new ArrayList<Printer>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see framework.Economy#getMarketFor(java.lang.Class, java.lang.Class)
     */
    public Market getMarketFor(String currencyName, String productName) {
        Market m = this.markets.get(this.getKey(currencyName, productName));
        if (m.getCurrency().equals(currencyName)) // check that it's not the
            // reverse Market.
            return m;
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see framework.Economy#run(int)
     */
    public void run(int steps) {
        for (int i = 0; i < steps; i++)
            step();
    }

    /*
     * (non-Javadoc)
     * 
     * @see framework.Economy#step()
     */
    public void step() {
        // have each Agent act, store in deadAgents if dead
        for (Agent a : this.agents) {
            a.act();
            if (!a.isAlive())
                this.deadAgents.add(a);
        }
        // clear all Markets
        for (Market m : this.markets.values())
            m.clear();
        // remove all dead Agents from the simulation
        for (Agent a : this.deadAgents)
            this.agents.remove(a);

        for (Agent a : this.deadAgents) {
            String s = "";
            if (a instanceof SingleOwnerFirmOwner) {
                SingleOwnerFirmOwner b = (SingleOwnerFirmOwner) a;
                if (b.getFirms() == null)
                    s = " (laborer) ";
                else
                    s = " (owner) ";
            }
            System.err
                    .println(a.getClass().getSimpleName() + s + " #"
                            + a.getID() + " died at step "
                            + this.runningTimeSteps + " with "
                            + a.getGood("Money").getQuantity() + " money and "
                            + a.getGood("Food").getQuantity() + " food");
        }
        if (this.deadAgents.size() > 0)
            System.err.println("Now there are " + this.agents.size() + " agents");

        this.deadAgents.clear();

        // have all Printers print
        for (Printer p : this.printers)
            p.print(this);
        
        if ((this.runningTimeSteps & 511) == 0)
        System.out.println(this.runningTimeSteps);
        
        // increment runningTimeSteps
        runningTimeSteps++;
    }

    /*
     * (non-Javadoc)
     * 
     * @see output.DataOutputEconomy#addPrinter(output.Printer)
     */
    public void addPrinter(Printer printer) {
        this.printers.add(printer);
        printer.printTitles(this);
    }

    /*
     * (non-Javadoc)
     * 
     * @see output.DataOutputEconomy#getAgents()
     */
    public Collection<Agent> getAgents() {
        return this.agents;
    }

    /*
     * (non-Javadoc)
     * 
     * @see output.DataOutputEconomy#getMarkets()
     */
    public Collection<Market> getMarkets() {
        return this.markets.values();
    }

    /*
     * (non-Javadoc)
     * 
     * @see output.DataOutputEconomy#getRunningTimeSteps()
     */
    public int getRunningTimeSteps() {
        return this.runningTimeSteps;
    }

    /**
     * Add agent to this Economy.
     */
    public void addAgent(Agent agent) {
        if (agent == null)
            err("market is null");
        this.agents.add(agent);
    }

    /**
     * Add market to this Economy.
     */
    public void addMarket(Market market) {
        if (market == null)
            err("market is null");
        String key = this.getKey(market.getCurrency(), market.getProduct());
        if (this.markets.containsKey(key))
            err("Economy already contains a Market for " + market.getCurrency()
                    + " and " + market.getProduct());
        this.markets.put(key, market);
    }

    /**
     * Add all agents to this Economy.
     */
    public void addAgents(Collection<? extends Agent> agents) {
        if (this.runningTimeSteps > 0)
            err("Simulation has started");
        this.agents.addAll(agents);
    }

    /**
     * Add all markets to this Economy.
     */
    public void addMarkets(Collection<? extends Market> markets) {
        if (this.runningTimeSteps > 0)
            err("Simulation has started");
        for (Market m : markets)
            this.addMarket(m);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataOutputEconomy#cleanupPrinters()
     */
    public void cleanupPrinters() {
        for (Printer printer : this.printers)
            printer.cleanup(this);
    }

    /*----------------- Private Methods ------------------*/

    /**
     * Always returns the same key string for any ordered input of s1 and s2.
     */
    private String getKey(String s1, String s2) {
        int seperator = s1.length() + s2.length();
        if (s1.compareTo(s2) < 0)
            return s1 + seperator + s2;
        else
            return s2 + seperator + s1;
    }
}
