/**
 * edit 4-2-09 Michael Adelson: Added a basic implementation of toString that
 * returns "(Product) for (Currency) Market". edit 10-7-09 Michael Adelson:
 * Added basic implementations of new isCurrency, isProduct, getCurrency, and
 * getProduct types. 
 * edit 10-17-09 Michael Adelson changed signature of offer() methods to take 
 * an Agent instead of a Good. This should give Markets more flexibility (to be 
 * used with contracts for example) going forward. Removed the isCurrency and 
 * isProduct methods, since these are no longer useful enough to merit framework
 * inclusion.
 * edit 10-28-09 Michael Adelson added Economy to default constructor so that 
 * Market's requiring randomization can use Economy.getRandom(). This addition
 * may also be useful if the Economy implements regulations or other sorts of 
 * Market control.
 * edit 10-28-09 Michael Adelson changed Market to use new Good getName() 
 * convention.
 */
package eos.framework;

/**
 * A Market is an object that enables the exchange of two Goods. Each Market's
 * currency and product Goods are specified in the constructor. Classes which
 * extend Market can and likely will differ in the precise mechanism used to
 * process and clear bids and offers. For instance, the baseline implementation
 * uses a call auction.
 */
public abstract class Market {

    /**
     * The name that defines the Market's currency.
     */
    protected final String currencyName;

    /**
     * The name that defines the Market's product.
     */
    protected final String productName;

    /**
     * Construct a Market using Currency and Product. Note that this constructor
     * does nothing with economy. It is up to classes that extend Market to
     * decide if they want to save economy.
     */
    public Market(Economy economy, String currencyName, String productName) {
        this.currencyName = currencyName;
        this.productName = productName;
    }

    /**
     * Add an offer to sell. The return value should be used to communicate
     * failure due to insufficient quantity. If an input that is illegal in all
     * cases is passed, however, Market should throw an exception.
     */
    public abstract boolean addSellOffer(Agent seller, double maxQuantity,
            double price);

    /**
     * Add an offer to buy. The return value should be used to communicate
     * failure due to insufficient quantity. If an input that is illegal in all
     * cases is passed, however, Market should throw an exception.
     */
    public abstract boolean addBuyOffer(Agent buyer, double maxQuantity,
            double price);

    /**
     * Clear the Market, performing all transactions according to the
     * implementation.
     */
    public abstract void clear();

    /**
     * Returns a double representation of the last "Market Price" defined
     * according to the implementation.
     */
    public abstract double getLastMarketPrice();

    /**
     * Returns a double representation of the total quantity of product sold the
     * last time clear() was called for this Market.
     */
    public abstract double getLastMarketVolume();

    /**
     * Get this Market's currency.
     */
    public String getCurrency() {
        return this.currencyName;
    }

    /**
     * Get this Market's product.
     */
    public String getProduct() {
        return this.productName;
    }

    /**
     * Returns a String representation of the Market.
     */
    public String toString() {
        return this.productName + " for " + this.currencyName
                + " Market\nPrice = " + this.getLastMarketPrice()
                + "\nVolume = " + this.getLastMarketVolume() + "\n";
    }

}
