/**
 * 
 */
package eos.io;

import eos.utilities.*;
import eos.framework.*;

/**
 * A DataPrinter class for tracking the status of individual Agents. This
 * implementation is not multi-run compatible.
 * 
 * @author Michael Adelson
 */
public class AgentTrackerDataPrinter implements DataPrinter {

    /**
     * The modes in which the DataPrinter can run.
     * 
     * @author Michael Adelson
     */
    public enum AgentTrackerMode {
        Good, Goods, NetWorth, ToString;
    }

    /**
     * The Agent to be tracked.
     */
    private Agent agent;

    /**
     * This DataPrinter's mode
     */
    private final AgentTrackerMode mode;

    /**
     * True or false depending on mode
     */
    private boolean isCSVCompatible;

    /**
     * Used in Good mode.
     */
    private String name;

    /**
     * Index of Agent.
     */
    private int index;

    /**
     * Class of Agent, if passed
     */
    private Class<? extends Agent> agentClass;

    /**
     * Construct a tracker for mode. If running in "Good" mode, arg must be the
     * String name of the tracked Good. For the index'th Agent in economy.
     */
    public AgentTrackerDataPrinter(int index, DataOutputEconomy economy,
            AgentTrackerMode mode, Object arg) {
        this.mode = mode;
        this.SetUpForMode(null, index, Agent.class, economy, arg);
    }

    /**
     * Construct a tracker for mode. If running in "Good" mode, arg must be the
     * String name of the tracked Good. This tracker is for agent.
     */
    public AgentTrackerDataPrinter(Agent agent, DataOutputEconomy economy,
            AgentTrackerMode mode, Object arg) {
        this.mode = mode;
        this.SetUpForMode(agent, -1, null, economy, arg);
    }

    /**
     * Construct a tracker for mode. If running in "Good" mode, arg must be the
     * String name of the tracked Good. For the index'th Agent of class
     * agentClass in economy.
     */
    public AgentTrackerDataPrinter(Class<? extends Agent> agentClass,
            int index, DataOutputEconomy economy, AgentTrackerMode mode,
            Object arg) {
        this.mode = mode;
        this.SetUpForMode(null, index, agentClass, economy, arg);
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataPrinter#getCleanupData(eos.io.DataOutputEconomy)
     */
    public String getCleanupData(DataOutputEconomy economy) {
        String data = this.getData(economy);

        // free for garbage collection
        this.agent = null;

        return data;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataPrinter#getData(eos.io.DataOutputEconomy)
     */
    public String getData(DataOutputEconomy economy) {
        if (this.agent == null) // try to find Agent
            this.findAgent(economy);
        if (this.agent == null) // if Agent not found...
            return "0"; // no data

        // get data by mode
        switch (this.mode) {
        case Good:
            return this.agent.getGood(this.name).getQuantity() + "";
        case Goods:
            String s = "";
            for (Good g : this.agent.getGoods())
                s += g.getName() + ":\t" + g.getQuantity() + "\n";
            return s;
        case NetWorth:
            return this.agent.getNetWorth() + "";
        case ToString:
            return this.agent.toString();
        default:
            return "Unsupported Mode " + this.mode;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataPrinter#getTitle(eos.io.DataOutputEconomy)
     */
    public String getTitle(DataOutputEconomy economy) {
        // create the identifier for the Agent based on available info
        String agentID = "";
        if (this.index == -1) // not located by index
            agentID += Strings.getShortNames(this.agent.getClass()) + " "
                    + this.agent.getID();
        else { // located by index
            agentID += Strings.getOrdinal(this.index) + " "
                    + Strings.getShortNames(this.agentClass) + " ";
            if (this.agent == null) // couldn't find the Agent
                agentID += "(Not initially in economy)";
            else if (this.agentClass != this.agent.getClass())
                agentID += "(" + Strings.getShortNames(this.agent.getClass())
                        + ")";
        }

        // get title by mode
        switch (this.mode) {
        case Good:
            return agentID + "'s " + this.name;
        case ToString:
            return agentID;
        default:
            return agentID + "'s " + this.mode.toString();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataPrinter#isCSVCompatible()
     */
    public boolean isCSVCompatible() {
        return this.isCSVCompatible;
    }

    /*
     * (non-Javadoc)
     * 
     * @see eos.io.DataPrinter#isMultiRunCompatible()
     */
    public boolean isMultiRunCompatible() {
        return false;
    }

    /*---------------------- PRIVATE METHODS --------------------------*/

    /**
     * Constructor helper.
     */
    private void SetUpForMode(Agent agent, int index,
            Class<? extends Agent> agentClass, DataOutputEconomy economy,
            Object arg) {
        // get the Agent
        if (agent != null) { // directly set the Agent
            this.agent = agent;
            this.index = -1; // not needed
            this.agentClass = null; // not needed
        } else { // get the index'th agent of agentClass
            this.index = index;
            this.agentClass = agentClass;
            this.findAgent(economy);
        }

        // select what to track
        switch (this.mode) {
        case Good:
            if (!(arg instanceof String))
                Errors
                        .err("Parameter arg must be the String name of the desired Good");
            this.name = String.class.cast(arg);
            this.isCSVCompatible = true;
            break;
        case NetWorth:
            this.name = null;
            this.isCSVCompatible = true;
            break;
        case Goods: // fall through
        case ToString:
            this.name = null;
            this.isCSVCompatible = false;
            break;
        default:
            Errors.err("Mode " + this.mode + " not supported");
        }
    }

    /**
     * Try to find the Agent based on index and agentClass.
     */
    private void findAgent(DataOutputEconomy economy) {
        int count = 0;
        for (Agent a : economy.getAgents()) {
            if (!this.agentClass.isInstance(a))
                continue;
            if (count == this.index) {
                this.agent = a;
                break;
            }
            count++;
        }
    }
}
