package eos.io;

import java.util.*;
import java.io.*;
import eos.utilities.Errors;

/**
 * Allows the output from multiple Economies (simulation runs) to be Printed
 * side by side by a single Printer object to a single output file.
 * 
 * @author Michael Adelson
 */
public class CSVMultiRunPrinter extends CSVPrinter implements MultiRunPrinter {

    /**
     * Strings used in naming temp files.
     */
    private static final String TEMP_NAME = "csvMultiRun_";
    private static final String FILE_EXTENSION = ".csv";

    /**
     * Class used to map Economies to printers and tempFiles.
     */
    private class PrinterAndFile {
        final File tempFile;
        final Printer printer;

        public PrinterAndFile(File tempFile, Printer printer) {
            this.tempFile = tempFile;
            this.printer = printer;
        }
    }

    /**
     * Number used in naming temp files.
     */
    protected static int tempNumber = 0;

    /**
     * Map of economies to their corresponding output objects.
     */
    protected LinkedHashMap<DataOutputEconomy, PrinterAndFile> map;

    /**
     * Old pafs (cleanup has already been called for their economies)
     */
    protected List<PrinterAndFile> oldPafs;

    /**
     * Construct a CSVMultiRunPrinter, starting with economy.
     */
    public CSVMultiRunPrinter(String fileName, int period) {
        super(fileName, period);
        this.map = new LinkedHashMap<DataOutputEconomy, PrinterAndFile>();
        oldPafs = new ArrayList<PrinterAndFile>();
    }

    /**
     * If called with a used economy, cleans up temp file for that economy. If
     * called with null, concatenates temp files into a single output file.
     */
    public void cleanup(DataOutputEconomy economy) {
        if (economy != null) // cleanup for that economy
        {
            PrinterAndFile paf;
            if (map.containsKey(economy)) // seen before
            {
                paf = map.get(economy);
                oldPafs.add(paf); // add to old
            } else // never seen before
            {
                paf = this.registerEconomy(economy);
            }
            paf.printer.cleanup(economy); // cleanup
            map.remove(economy); // remove map to allow economy to be gc'd
        } else // perform concatenation
        {
            this.cleanup();
        }
    }

    /**
     * Equivalent to cleanup(null); concatenates temp files into a single file.
     */
    public void cleanup() {
        // have all unfinished printers cleanup
        for (DataOutputEconomy doe : this.map.keySet()) {
            PrinterAndFile paf = this.map.get(doe);
            paf.printer.cleanup(doe);
        }

        // read in and concatenate tempfiles

        // combine old and new pafs
        this.oldPafs.addAll(this.map.values());
        this.map = null;

        // create readers
        List<BufferedReader> readers = new ArrayList<BufferedReader>();
        for (PrinterAndFile paf : this.oldPafs) {
            try {
                readers.add(new BufferedReader(new FileReader(paf.tempFile)));
            } catch (Exception e) {
                Errors.err(e.getMessage());
            }
        }

        // write Economy headers
        String line = null;
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < this.oldPafs.size(); i++) {
            sb.append("Run " + i + ": " + CSVPrinter.CS);
            // column titles
            try {
                line = readers.get(i).readLine();
            } catch (Exception e) {
                Errors.err(e.getMessage());
            }
            if (line == null || line.equals(""))
                Errors.err("No data for Run " + i);
            sb.append(line);
            // leave a column between runs
            sb.append(CSVPrinter.CS);
        }
        this.printWriter.println(sb.toString());

        // read and concatenate lines
        boolean doneReading = false;
        while (!doneReading) {
            doneReading = true;
            sb.setLength(0);
            for (BufferedReader reader : readers) {
                // leave a column for the Run label
                sb.append(CSVPrinter.CS);
                try {
                    line = reader.readLine();
                } catch (Exception e) {
                    Errors.err(e.getMessage());
                }

                if (line != null && !line.equals("")) {
                    doneReading = false;
                    sb.append(line);
                } else // otherwise pad with empty columns
                {
                    for (int i = 0; i < this.dataPrinters.size() + 1; i++)
                        sb.append(CSVPrinter.CS);
                }
                // leave a column between runs
                sb.append(CSVPrinter.CS);
            }
            this.printWriter.println(sb.toString());
        }

        // close readers
        for (BufferedReader reader : readers) {
            try {
                reader.close();
            } catch (IOException e) {
                Errors.err(e.getMessage());
            }
        }

        // delete temp files
        for (PrinterAndFile paf : this.oldPafs) {
            paf.tempFile.delete();
        }

        // close printWriter
        this.printWriter.close();
    }

    /**
     * Prints for economy if applicable. If economy has not been seen before,
     * CSVMultiRunPrinter registers a new "run" for this Economy.
     */
    public void print(DataOutputEconomy economy) {
        PrinterAndFile paf = map.get(economy);

        // register a new run if necessary
        if (paf == null)
            paf = registerEconomy(economy);

        // print
        paf.printer.print(economy);
    }

    /**
     * Calls printTitles() for the printer corresponding to economy, or creates
     * a new run for economy if necessary.
     */
    public void printTitles(DataOutputEconomy economy) {
        PrinterAndFile paf = map.get(economy);

        // register a new run as necessary
        if (paf == null)
            paf = registerEconomy(economy);

        // title
        paf.printer.printTitles(economy);
    }

    /**
     * Add a DataPrinter to this Printer. This should be called before any of
     * the economies are run.
     */
    public void addDataPrinter(DataPrinter dataPrinter) {
        if (!dataPrinter.isMultiRunCompatible())
            Errors.err("DataPrinter " + dataPrinter
                    + " is not multiRun compatible");
        super.addDataPrinter(dataPrinter);
        for (PrinterAndFile paf : this.map.values()) {
            paf.printer.addDataPrinter(dataPrinter);
        }
    }

    /**
     * Obtain a new temp file.
     */
    private File getTempFile() {
        String tempFileName;
        do {
            tempFileName = TEMP_NAME + tempNumber + FILE_EXTENSION;
            tempNumber++;
        } while (Printers.hasPrintWriterFor(tempFileName));

        return new File(tempFileName);
    }

    /**
     * Adds a mapping with economy to map.
     */
    private PrinterAndFile registerEconomy(DataOutputEconomy economy) {
        File file = getTempFile();
        CSVPrinter printer = new CSVPrinter(file.getName(), this.period);
        for (DataPrinter dataPrinter : this.dataPrinters)
            printer.addDataPrinter(dataPrinter);
        PrinterAndFile paf = new PrinterAndFile(file, printer);
        map.put(economy, paf);
        return paf;
    }

}
