package eos.io;

import java.util.*;
import eos.framework.Market;

/**
 * A DataPrinter for Market volumes.
 * 
 * @author Michael Adelson
 */
public class MarketVolumeDataPrinter implements DataPrinter {

    private final String currencyName, productName;
    
    /**
     * Defines the default currency String used.
     */
    public static final String DEFAULT_CURRENCY = "Money";
    
    /**
     * Assumes currency is DEFAULT_CURRENCY.
     */
    public MarketVolumeDataPrinter(String productName)
    {
        this.productName = productName;
        this.currencyName = DEFAULT_CURRENCY;
    }
    
    /**
     * Create a MarketPriceDataPrinter for currency and product. 
     */
    public MarketVolumeDataPrinter(String currencyName, 
                                  String productName)
    {
        this.productName = productName;
        this.currencyName = currencyName;
    }
    
    /**
     * Calls this.getData(economy)
     */
    public String getCleanupData(DataOutputEconomy economy) {
        return this.getData(economy);
    }

    /**
     * Returns the last MarketPrice, or a "Not Found" statement.
     */
    public String getData(DataOutputEconomy economy) {
        Market m = economy.getMarketFor(currencyName, productName);
        if (m == null)
            return "No Market Found";
        else 
            return m.getLastMarketVolume() + "";
    }

    /* (non-Javadoc)
     * @see output.DataPrinter#getTitle(output.DataOutputEconomy)
     */
    public String getTitle(DataOutputEconomy economy) {
        if (this.currencyName.equals(DEFAULT_CURRENCY))
            return this.productName + " volume";
        else 
            return this.productName + " volume (" + this.currencyName + ")";
    }

    /**
     * True.
     */
    public boolean isCSVCompatible() {
        return true;
    }

    /**
     * True.
     */
    public boolean isMultiRunCompatible() {
        return true;
    }
    
    /**
     * Create a MarketPriceDataPrinter for each Market in economy.
     */
    public static Collection<DataPrinter> getDataPrintersFor(DataOutputEconomy economy)
    {
        List<DataPrinter> list = new ArrayList<DataPrinter>();
        for (Market m : economy.getMarkets())
        {
            if (m.getCurrency().equals(DEFAULT_CURRENCY))
                list.add(new MarketVolumeDataPrinter(m.getProduct()));
            else
                list.add(new MarketVolumeDataPrinter(m.getCurrency(), m.getProduct()));
        }
        
        return list;
    }
}
