package eos.io;

import java.util.*;

import eos.utilities.Strings;
import eos.framework.*;

/**
 * Can be used to track the net worth of Agents. The output can be average or
 * total, and can be tracked by class, by individual, or by total population.
 * 
 * @author Michael Adelson
 */
public class NetWorthDataPrinter implements DataPrinter {

    private final boolean avg;
    private final Collection<Class<? extends Agent>> classes;
    private final int agentID;

    /**
     * By population. Can be set to print AVERAGE, rather than TOTAL, 
     * net worth.
     */
    public NetWorthDataPrinter(boolean printAvg) {
        this.avg = printAvg;
        this.classes = null;
        this.agentID = -1;
    }

    /**
     * By classes. Can be set to print AVERAGE, rather than TOTAL, 
     * net worth.
     */
    public NetWorthDataPrinter(Collection<Class<? extends Agent>> classes,
            boolean printAvg) {
        this.avg = printAvg;
        this.classes = classes;
        this.agentID = -1;
    }

    /**
     * By class
     */
    public NetWorthDataPrinter(Class<? extends Agent> agentClass,
            boolean printAvg) {
        Collection<Class<? extends Agent>> classes = new ArrayList<Class<? extends Agent>>();
        classes.add(agentClass);
        this.avg = printAvg;
        this.classes = classes;
        this.agentID = -1;
    }

    /**
     * By individual
     */
    public NetWorthDataPrinter(int agentID) {
        this.avg = false;
        this.classes = null;
        this.agentID = agentID;
    }

    /**
     * Calls this.getData(economy)
     */
    public String getCleanupData(DataOutputEconomy economy) {
        return this.getData(economy);
    }

    /**
     * Returns data based on constructor parameters.
     */
    public String getData(DataOutputEconomy economy) {
        int count = 0;
        double netWorth = 0.0;
        if (this.classes != null) // by class(es)
        {
            for (Agent a : economy.getAgents()) {
                if (!this.classes.contains(a.getClass()))
                    continue;
                count++;
                netWorth += a.getNetWorth();
            }
        } else if (this.agentID != -1) // by individual
        {
            for (Agent a : economy.getAgents()) {
                if (a.getID() == this.agentID) {
                    netWorth += a.getNetWorth();
                }
            }
        } else // by population
        {
            for (Agent a : economy.getAgents()) {
                count++;
                netWorth += a.getNetWorth();
            }
        }
        if (this.avg)
            return (netWorth / count) + "";
        return netWorth + "";
    }

    /**
     * Title based on constructor parameters
     */
    public String getTitle(DataOutputEconomy economy) {
        String title;
        if (this.avg)
            title = "Avg";
        else
            title = "Total";
        if (this.classes != null) {
            if (this.classes.size() == 1) {
                title += " NW of "
                        + Strings.getShortNames((Class<?>) this.classes
                                .toArray()[0]) + "s";
            } else {
                title += " NW of classes: ";
                for (Class<? extends Agent> cls : this.classes) {
                    title += Strings.getShortNames(cls) + "; ";
                }
            }
        } else if (this.agentID != -1) {
            title += "NW of Agent " + this.agentID;
        } else {
            title += "Net Worth";
        }
        return title;
    }

    /**
     * True.
     */
    public boolean isCSVCompatible() {
        return true;
    }

    /**
     * True.
     */
    public boolean isMultiRunCompatible() {
        return true;
    }

}
