package eos.io;

import java.io.*;
import java.util.*;
import eos.utilities.Errors;

/**
 * A printer object allows the printing of simulation data to an output file. It
 * maintains a collection of DataPrinters.
 * 
 * @author madelson
 */
public abstract class Printer {

    /**
     * The instance of PrintWriter used by this Printer.
     */
    protected PrintWriter printWriter;
    
    /**
     * The time step interval for printing.
     */
    protected final int period;
    
    /**
     * The printer's fileName
     */
    protected final String fileName;
    
    /**
     * The list of DataPrinters used by this Printer.
     */
    protected List<DataPrinter> dataPrinters;
    
    /**
     * Base constructor.
     */
    public Printer(String fileName, int period)
    {
        this.printWriter = Printers.getPrintWriterFor(fileName);
        this.period = period;
        this.fileName = fileName;
        this.dataPrinters = new ArrayList<DataPrinter>();
    }
    
    /**
     * Get this Printer's fileName.
     */
    public String getFileName() {
        return this.fileName;
    }
    
    /**
     * Get this Printer's period.
     */
    public int getPrintPeriod() {
        return this.period;
    }
    
    /**
     * Add a DataPrinter to this Printer. Adding a DataPrinter in the middle of
     * a run may cause incorrect behavior.
     */
    public void addDataPrinter(DataPrinter dataPrinter) 
    {
        if (dataPrinter == null)
            Errors.err("Null DataPrinter passed to Printer.addDataPrinter");           
        dataPrinters.add(dataPrinter);
    }
    
    /**
     * Calls this.addDataPrinter for each dataPrinter in the collection.
     */ 
    public void addDataPrinters(Collection<DataPrinter> dataPrinters)
    {
        for (DataPrinter dp : dataPrinters)
            this.addDataPrinter(dp);
    }
    
    /**
     * This method should print Titles for each of this Printer's dataPrinters.
     * 
     */
    public abstract void printTitles(DataOutputEconomy economy);
    
    /**
     * This method should print the next piece of data for this Printer from
     * this economy. The printer's behavior is not defined if print is called
     * twice on the same timestep.
     */
    public abstract void print(DataOutputEconomy economy);
    
    /**
     * Performs any necessary cleanup. This method should always be called when
     * the simulation has completed (or if errors occur).
     */
    public abstract void cleanup(DataOutputEconomy economy);  
}
