package eos.io;

import java.io.*;
import java.util.*;
import eos.utilities.*;

/**
 * Contains static methods used to coordinate Printer objects.
 * 
 * @author Michael Adelson
 */
public final class Printers {

    /**
     * Holds all PrintWriters, allowing lookup by filename.
     */
    private static final Map<String, PrintWriter> printWriters;
    
    /**
     * Initializer
     */
    static {
        printWriters = new HashMap<String, PrintWriter>();
    }
    
    /**
     * Not instantiable.
     */
    private Printers() {}
    
    /**
     * Returns a reference to the current printWriter for fileName, or creates
     * a new PrintWriter if none is available. Printers should always obtain
     * there writers via this method so that only one PrintWriter is open for 
     * each file at a time. A static reference to the associated File object
     * is maintained. Throws a RuntimeException if the method fails to 
     * create a new PrintWriter.
     */
    public static synchronized PrintWriter getPrintWriterFor(String fileName)
    {
        if (printWriters.containsKey(fileName))
            return printWriters.get(fileName);
        
        PrintWriter printWriter = null;
        try {
            printWriter = new PrintWriter(fileName);
            printWriters.put(fileName, printWriter);
        } catch (Exception e) {
            Errors.err("Failed to create " + fileName + ": check that " + fileName + " is not open");
        }
        
        return printWriter;
    }
    
    /**
     * Returns true if a PrintWriter has been created for fileName.
     */
    public static boolean hasPrintWriterFor(String fileName)
    {
        return printWriters.containsKey(fileName);
    }
}
