/**
 * 
 */
package eos.io;

import java.io.PrintWriter;
import java.util.*;
import eos.framework.*;
import eos.models.SingleOwnerFirms.*;
import eos.utilities.Errors;
import eos.utilities.Simulations;
import eos.utilities.Strings;
import eos.utilities.Simulations.*;

/**
 * A class that helps simulation-package classes efficiently run and construct
 * simulations. Supports SingleOwnerFirms.
 * 
 * @author Michael Adelson
 */
public class Simulation {

    /**
     * The default Economy class.
     */
    private Class<? extends Economy> defaultEconomyClass = null;

    /**
     * The simulation's name
     */
    private final String name;

    /**
     * @param name
     *            the Simulation's name
     * @param economyClass
     *            the default economy class used to construct economies.
     * @param seed
     *            the seed used for the Simulation's Random(). Used to generate
     *            seeds for economy random numbers.
     */
    public Simulation(String name, Class<? extends Economy> economyClass,
            long seed) {
        this.name = name;

        this.defaultFileName = ((Strings.isNoE(this.name)) ? "sim" : this.name)
                + "_out%s.%s";

        this.defaultEconomyClass = economyClass;
        // seed rand
        this.rand.setSeed(seed);

        // setup log
        String logFileName = this.name + logFileSuffix;
        Errors.Assert(!Printers.hasPrintWriterFor(logFileName),
                "Can't create log file because there is already a PrintWriter for "
                        + logFileName);
        this.logPrinter = Printers.getPrintWriterFor(logFileName);
        this.logln("Simulation seed set to: " + seed);
    }

    /*
     * Methods relating to Randomization.
     */

    /**
     * Used to generate economy seeds.
     */
    private final Random rand = new Random();

    /**
     * Random seeds used.
     */
    private final List<Long> seeds = new ArrayList<Long>();

    /**
     * Returns this Simulation's random number generator.
     */
    public Random getRandom() {
        return this.rand;
    }

    /**
     * Get the Random seeds generated for Economies. Returns 1 random seed for
     * each economy instantiated.
     */
    public List<Long> getRandomSeeds() {
        return this.seeds;
    }

    /**
     * Get a new seed and record it in seeds.
     */
    private long getSeed() {
        long seed = this.rand.nextLong();
        this.seeds.add(seed);

        return seed;
    }

    /*
     * Initializing The Economy: These methods add Agents, Markets, and Goods to
     * the Simulation.
     */

    /**
     * Represents a Market to be instantiated
     */
    private class MarketInfo {
        Class<? extends Market> cls;
        String currencyName, productName;

        /**
         * @return An instantiation using this MarketInfo.
         */
        Market instantiate(Economy economy) {
            logln("Instantiated Market: class=%s, product=%s, currency=%s",
                    this.cls.getSimpleName(), this.productName,
                    this.currencyName);

            return Simulations.createMarket(this.cls, this.currencyName,
                    this.productName, economy);
        }
    }

    /**
     * Represents a group of Agents to be instantiated.
     */
    private class AgentInfo {
        Class<? extends Agent> cls;
        int pop;

        /**
         * @return An instantiation using this AgentInfo
         */
        List<? extends Agent> instantiate(Economy economy) {
            logln("Instantiated %d %ss", this.pop, this.cls.getSimpleName());

            Errors.Assert(startingQuantities.containsKey(this.cls),
                    "No starting quantities for: " + this.cls.getSimpleName());
            return Simulations.createAgents(this.cls, goods, startingQuantities
                    .get(this.cls), economy, this.pop);
        }
    }

    /**
     * Default goods map
     */
    private NameClsMap goods = null;

    /**
     * A map of Agent classes to their starting quantities
     */
    private final Map<Class<? extends Agent>, double[]> startingQuantities = new HashMap<Class<? extends Agent>, double[]>();

    /**
     * What percent to perturb each Agent's startingQuantities
     */
    private int initialPerturbation = 0;

    /**
     * Defines the identifiers used.
     */
    private final Map<String, List<Class<? extends Agent>>> identifiers = new HashMap<String, List<Class<? extends Agent>>>();

    /**
     * The MarketInfos to instantiate
     */
    private final List<MarketInfo> marketInfos = new ArrayList<MarketInfo>();

    /**
     * The AgentInfos to instantiate
     */
    private final List<AgentInfo> agentInfos = new ArrayList<AgentInfo>();

    /**
     * Specify markets in the simulation
     * 
     * @param marketClass
     *            the class of the Market
     * @param currencyName
     *            the currency to be used
     * @param productNames
     *            the products to be used. For each productName, creates a
     *            Market for currency, product.
     */
    public void addMarkets(Class<? extends Market> marketClass,
            String currencyName, String... productNames) {
        Errors.Assert(productNames.length != 0, "Must list at least 1 product");

        for (String productName : productNames) {
            MarketInfo mi = new MarketInfo();
            mi.cls = marketClass;
            mi.currencyName = currencyName;
            mi.productName = productName;
            this.marketInfos.add(mi);

            this.logln("Adding Market: Class: %s, product: %s, currency: %s",
                    marketClass.getSimpleName(), productName, currencyName);
        }
    }

    /**
     * Add default Goods instantiated with goodClass and possessing the names in
     * names.
     */
    public void addGoods(Class<? extends Good> goodClass, String... names) {
        String s = Arrays.toString(names);
        this.logln("Adding Goods: Class: %s, Names: %s", goodClass
                .getSimpleName(), s.substring(1, s.length() - 1));

        if (this.goods == null) {
            this.goods = Simulations.getMap(goodClass, names);
        } else
            for (String name : names)
                Errors.warnIf(this.goods.put(name, goodClass) != null,
                        "Overwriting existing mapping for " + name);
    }

    /**
     * Set the starting quantities of default Goods for Agents of class
     * agentClass. The first double in startinQuantities is the starting
     * quantity for the first default Good added to the simulation, and so on.
     * 
     * @param agentClass
     *            the class of the Agent to receive startingQuantities
     * @param startingQuantities
     *            the starting quantities of various Goods
     */
    public void setQuantities(Class<? extends Agent> agentClass,
            double... startingQuantities) {
        this.logln("Setting Starting Quantites: %s for %s", Arrays
                .toString(startingQuantities), agentClass);

        Errors
                .warnIf(startingQuantities.length != this.goods.size(),
                        "Number of starting quantities listed does not match the number of Goods");

        Errors.warnIf(this.startingQuantities.put(agentClass,
                startingQuantities) != null,
                "Overwriting existing mapping for " + agentClass);
    }

    /**
     * @param percentage
     *            set the max amount by which starting quantities are initially
     *            perturbed to be percentage percent
     */
    public void setPerturbation(int percentage) {
        Errors.Assert(percentage >= 0);

        this.initialPerturbation = percentage;

        this.logln("Set initial perturbation to %d percent", percentage);
    }

    /**
     * For each class that identifier represents, set the starting quantities of
     * default Goods for Agents of that class. The first double in
     * startinQuantities is the starting quantity for the first default Good
     * added to the simulation, and so on.
     * 
     * @param identifier
     *            the classes of Agent to receive startingQuantities. Set with
     *            createIdentifier().
     * @param startingQuantities
     *            the starting quantities of various Goods
     */
    public void setQuantities(String identifier, double... startingQuantities) {
        String id = identifier.toLowerCase();
        Errors.Assert(this.identifiers.containsKey(id), "Identififer "
                + identifier + " not found");

        String s = "";
        for (Class<? extends Agent> cls : this.identifiers.get(id)) {
            this.setQuantities(cls, startingQuantities);
            s += cls.getSimpleName() + ", "; // used for logging
        }

        this.logln("Setting Starting Quantities: %s for %s (%s)", Arrays
                .toString(startingQuantities), id, s.substring(0,
                s.length() - 2));
    }

    /**
     * Creates a lower case identifier for the list of classes. Some methods of
     * this class have overloads that take identifiers rather than the classes
     * themselves.
     * 
     * @param identifier
     *            the identifier String to be used
     * @param agentClasses
     *            the classes of Agents to identify
     * @return the actual identifier used (all lower case)
     */
    @SuppressWarnings("unchecked")
    public String createIdentifier(String identifier, Class<?>... agentClasses) {
        String id = identifier.toLowerCase();
        Errors.warnIf(this.identifiers.containsKey(id),
                "Overwriting identifier " + id);

        this.logln("Creating Identifier \"%s\" for %s", id, Strings
                .getShortNames(agentClasses));

        List<Class<? extends Agent>> list = new ArrayList<Class<? extends Agent>>();
        for (Class<?> agentClass : agentClasses)
            list.add((Class<? extends Agent>) agentClass);

        this.identifiers.put(id, list);

        return id;
    }

    /**
     * Add pop Agents to the simulation. Does nothing if pop <= 0.
     */
    public void addAgents(Class<? extends Agent> agentClass, int pop) {
        Errors.warnIf(pop <= 0, "Can't add " + pop + " "
                + agentClass.getSimpleName() + "s. Ignoring addAgents call.");
        if (pop <= 0)
            return;

        this.logln("Added %d %ss", pop, agentClass.getSimpleName());

        AgentInfo ai = new AgentInfo();
        ai.cls = agentClass;
        ai.pop = pop;

        this.agentInfos.add(ai);
    }

    /**
     * A shortcut method to support SingleOwnerFirm ownership. Adds pop owners
     * to the simulation as well as pop * numFirms firms. Each owner initially
     * owns numFirms Firms. Assumes that the Economy implements
     * DataOutputEconomy.
     * 
     * @param <O>
     *            the Owner class
     * @param <F>
     *            the Firm class
     * @param ownerClass
     *            the Owner class
     * @param pop
     *            the number of owners to add
     * @param firmClass
     *            the Firm class
     * @param numOwned
     *            the number of firms per owner
     */
    public <O extends Agent & SingleOwnerFirmOwner, F extends Agent & SingleOwnerFirm> void addOwnersAndFirms(
            final Class<O> ownerClass, final int pop, final Class<F> firmClass,
            final int numOwned) {
        this.addAgents(ownerClass, pop);
        this.addAgents(firmClass, pop * numOwned);

        this.logln("Assigned each of the first %d %ss ownership of %d %ss",
                pop, ownerClass.getSimpleName(), numOwned, firmClass
                        .getSimpleName());

        InitializationFunction iFunc = new InitializationFunction() {
            final int ownerPop = pop, firmPop = pop * numOwned;
            final Class<O> ocls = ownerClass;
            final Class<F> fcls = firmClass;

            public void initialize(Economy economy) {
                Errors.Assert(economy instanceof DataOutputEconomy);
                DataOutputEconomy doe = (DataOutputEconomy) economy;

                SingleOwnerFirmOwner[] owners = new SingleOwnerFirmOwner[this.ownerPop];
                SingleOwnerFirm[] firms = new SingleOwnerFirm[this.firmPop];
                int ownerCount = 0, firmCount = 0;
                for (Agent agent : doe.getAgents()) {
                    // if an owner with no firms...
                    if (ocls.isAssignableFrom(agent.getClass())
                            && ownerCount < owners.length) {
                        SingleOwnerFirmOwner owner = (SingleOwnerFirmOwner) agent;
                        if (owner.getFirms() == null
                                || owner.getFirms().size() == 0) {
                            owners[ownerCount] = (SingleOwnerFirmOwner) agent;
                            ownerCount++;
                        }
                    } // if a firm with no owner
                    else if (fcls.isAssignableFrom(agent.getClass())
                            && ((SingleOwnerFirm) agent).getOwner() == null
                            && firmCount < firms.length) {
                        firms[firmCount] = (SingleOwnerFirm) agent;
                        firmCount++;
                    }
                }

                // check that all owners and firms found
                Errors.Assert(ownerCount == this.ownerPop);
                Errors.Assert(firmCount == this.firmPop);

                // give each owner firmsPerOwner firms
                int firmsPerOwner = this.firmPop / this.ownerPop;
                firmCount = 0;
                for (SingleOwnerFirmOwner owner : owners) {
                    for (int i = 0; i < firmsPerOwner; i++, firmCount++)
                        firms[firmCount].transferOwnership(owner);
                }
            }

            // for logging purposes
            public String toString() {
                return "firm ownership assignment function";
            }
        };

        // add iFunc
        this.addInitializationFunction(iFunc);
    }

    /*
     * Initialization Functions: These functions allow the user to add
     * initialization functions which allow for specialized customization.
     */

    /**
     * A special functor interface that can be created anonymously by users to
     * perform special actions upon creating an economy. Initialize could, for
     * example, add special dataPrinters to economy.
     * 
     * @author Michael Adelson
     * 
     */
    public static interface InitializationFunction {
        void initialize(Economy economy);
    }

    /**
     * The InitializationFunctions to be run.
     */
    private final List<InitializationFunction> initFunctions = new ArrayList<InitializationFunction>();

    /**
     * Adds func to this Simulation's list of startup functions that are run
     * after economy initialization.
     */
    public void addInitializationFunction(InitializationFunction func) {
        this.logln("Adding InitializationFunction [%s]", func.toString());

        this.initFunctions.add(func);
    }

    /*
     * Running the Simulation: Methods to instantiate and run the simulation.
     */

    /**
     * The number of economies instantiated.
     */
    private int numEconomies = 0;

    /**
     * Get an initialized Economy of class economyClass.
     */
    public <E extends Economy> E getEconomy(Class<E> economyClass) {
        this.logln("Created Economy from Class: %s", economyClass
                .getSimpleName());

        E economy = Simulations.create(economyClass); // instantiate with no
        // arguments
        this.setupEconomy(economy);
        return economy;
    }

    /**
     * Get an initialized Economy of the default economyClass.
     */
    public Economy getEconomy() {
        return this.getEconomy(this.defaultEconomyClass);
    }

    /**
     * Run an economy of the default class for steps steps. Cleans up related
     * Printers.
     */
    public void run(int steps) {
        this.run(this.defaultEconomyClass, steps);
    }

    /**
     * Run an economyClass economy for steps steps. Cleans up related Printers.
     */
    public void run(Class<? extends Economy> economyClass, int steps) {
        Economy economy = this.getEconomy(economyClass);
        economy.run(steps);
        if (economy instanceof DataOutputEconomy)
            ((DataOutputEconomy) economy).cleanupPrinters();

        this.logln("Ran economy for %d time steps", steps);
    }

    /**
     * Perform runs simulations of the default economy class with stepsPerRun
     * steps each.
     */
    public void run(int stepsPerRun, int runs) {
        this.run(this.defaultEconomyClass, stepsPerRun, runs);
    }

    /**
     * Perform runs simulations of the default economy class with stepsPerRun
     * steps each.
     */
    public void run(Class<? extends Economy> economyClass, int stepsPerRun,
            int runs) {
        for (int i = 0; i < runs; i++)
            this.run(economyClass, stepsPerRun);

        this.logln("Performed %d runs", runs);
    }

    /**
     * Sets up economy for this simulation.
     */
    private void setupEconomy(Economy economy) {
        // set seed
        long seed = this.getSeed();
        economy.getRandom().setSeed(seed);
        this.logln("Economy initialized with seed " + seed);
        this.numEconomies++;

        // add Markets
        for (MarketInfo mi : this.marketInfos)
            economy.addMarket(mi.instantiate(economy));

        // add Agents
        for (AgentInfo ai : this.agentInfos)
            for (Agent agent : ai.instantiate(economy))
                economy.addAgent(agent);

        // perform initialization functions
        for (InitializationFunction iFunc : this.initFunctions)
            iFunc.initialize(economy);

        // check if economy is a DOE
        DataOutputEconomy doe;
        if (economy instanceof DataOutputEconomy)
            doe = (DataOutputEconomy) economy;
        else {
            if (this.printingEnabled || this.initialPerturbation > 0)
                this
                        .logln("Economy is not a DataOutputEconomy: can't perform initial perturbations or add Printers");
            return;
        }

        // perform initial perturbations
        double maxPercent = (double) this.initialPerturbation / 100.0;
        for (Agent agent : doe.getAgents()) {
            for (Good good : agent.getGoods()) {
                if (good.isDiscrete()) // don't mess with discrete goods
                    continue;
                double amount = good.getQuantity() * maxPercent
                        * this.rand.nextDouble();
                if (this.rand.nextBoolean()) // randomly raise or lower
                    good.deposit(amount);
                else
                    good.withdraw(amount);
            }
        }

        // add Printers
        if (this.printingEnabled) {

            // default CSV
            if (this.defaultCSVEnabled) {
                String id = this.numEconomies == 0 ? "" : this.numEconomies
                        + "";
                CSVPrinter csvp = new CSVPrinter(this.getFileName(id),
                        this.defaultPrintPeriod);
                csvp.addDataPrinters(this.getCSVDataPrinters());
                // add user-added dataPrinters
                for (DataPrinter dp : this.dataPrinters.keySet()) {
                    if (this.dataPrinters.get(dp) == null)
                        csvp.addDataPrinter(dp);
                }
                doe.addPrinter(csvp);

                this.logln("Added Default CSV Printer for %s", csvp
                        .getFileName());
            }
            // default CSVMRP
            if (this.defaultMRP != null) {
                doe.addPrinter(this.defaultMRP);

                this.logln("Added Default MultiRunPrinter for %s",
                        this.defaultMRP.getFileName());
            }
            // default cleanup
            if (this.defaultCleanup != null) {
                doe.addPrinter(this.defaultCleanup);

                this.logln("Added Default Cleanup Printer for %s",
                        this.defaultCleanup.getFileName());
            }
            // user-added
            // remove all non-multirun printers and data printers
            List<Object> toRemove = new ArrayList<Object>();
            List<Printer> initPrinters = new ArrayList<Printer>();
            for (Printer p : this.myPrinters.keySet()) {
                this.logln("Adding Printer: class=%s, filename=%s, period=%d",
                        p.getClass().getSimpleName(), p.getFileName(), p
                                .getPrintPeriod());

                // check if already initialized
                if (!this.myPrinters.get(p)) { // initialize
                    initPrinters.add(p);

                    for (DataPrinter dp : this.dataPrinters.keySet()) {
                        if (this.dataPrinters.get(dp).equals(p.getFileName()))
                            p.addDataPrinter(dp);
                        if (!dp.isMultiRunCompatible())
                            toRemove.add(dp);
                    }
                }

                // add to the economy
                doe.addPrinter(p);

                // remove if not a multirun printer
                if (!(p instanceof MultiRunPrinter)) {
                    toRemove.add(p);
                }
            }

            // set all printers to initialized
            for (Printer p : initPrinters)
                this.myPrinters.put(p, true);

            // remove any non-multirun Printers and DataPrinters
            for (Object printer : toRemove) {
                this.logln("Removing IO object %s from future runs", printer
                        .toString());

                if (printer instanceof DataPrinter)
                    this.dataPrinters.remove(printer);
                else if (printer instanceof Printer)
                    this.myPrinters.remove(printer);
                else
                    Errors.Assert(false); // should never get here
            }
        }
    }

    /*
     * Printing: Methods to manage printing.
     */

    private boolean printingEnabled = true;
    private boolean loggingEnabled = true;
    private boolean defaultCSVEnabled = true;

    /**
     * The default multi-run printer
     */
    private CSVMultiRunPrinter defaultMRP = null;

    /**
     * The default cleanup printer
     */
    private CSVMultiRunCleanupPrinter defaultCleanup = null;

    /**
     * The user-added DataPrinters mapped to filenames. A linked map is used so
     * that the columns will be ordered in a regular fashion.
     */
    private Map<DataPrinter, String> dataPrinters = new LinkedHashMap<DataPrinter, String>();

    /**
     * User-added Printers. Maps Printers to values which are true if the
     * Printers have been initialized (this value is relevant for
     * MultiRunPrinters so that the DataPrinters won't get added twice).
     */
    private Map<Printer, Boolean> myPrinters = new HashMap<Printer, Boolean>();

    /**
     * The base file name used by Simulations. Set in constructor.
     */
    private String defaultFileName;

    /**
     * The default file extension used
     */
    private String defaultExtension = "csv";

    /**
     * The default printer period used
     */
    private int defaultPrintPeriod = 50;

    /**
     * The name of a simulation's log file is sim.name + logFileSuffix
     */
    private static String logFileSuffix = "_log.txt";

    /**
     * The log used for this Simulation. Initialized in the constructor.
     */
    private PrintWriter logPrinter = null;

    /**
     * Turn off all printing.
     */
    public void disablePrinting() {
        this.logln("Printing Disabled");

        this.printingEnabled = false;
    }

    /**
     * Enable all printing. Enabled as a default.
     */
    public void enablePrinting() {
        this.logln("Printing Enabled");

        this.printingEnabled = true;
    }

    /**
     * Disable printing "log" output to a logfile.
     */
    public void disableLogging() {
        this.loggingEnabled = false;
    }

    /**
     * Enable printing "log" output to a logfile.
     */
    public void enableLogging() {
        this.loggingEnabled = true;
    }

    /**
     * Log formatString formatted with args if logging is enabled.
     */
    public void log(String formatString, Object... args) {
        if (this.loggingEnabled) {
            this.logPrinter.print(String.format(formatString, args));
            this.logPrinter.flush();
        }
    }

    /**
     * Log formatString formatted with args and a newline character if logging
     * is enabled.
     */
    public void logln(String formatString, Object... args) {
        if (this.loggingEnabled) {
            this.logPrinter.println(String.format(formatString, args));
            this.logPrinter.flush();
        }
    }

    /**
     * Set the fileName used for default printing. Should be a format String
     * that accepts a run number and an extension such as: "sim_out%s.%s".
     */
    public void setDefaultFileName(String fileName) {
        try {
            String.format(fileName, 0, "csv"); // test
        } catch (Exception ex) {
            Errors.err(ex, "fileName did not conform to formatting standards");
        }

        this.defaultFileName = fileName;

        this.logln("Set default filename to "
                + String.format(fileName, "[Run Number]", "[Extension]"));
    }

    /**
     * Sets the base extension used by Simulation default Printers. Default is
     * "csv". NOTE that the '.' character is assumed and should not be included
     * in extension.
     */
    public void setDefaultExtension(String extension) {
        this.defaultExtension = extension.replace('.', ' ');

        this.logln("Set default file extension to ." + extension);
    }

    /**
     * Set the default print period used to instantiate future printers. Default
     * is 50.
     */
    public void setDefaultPrintPeriod(int period) {
        this.defaultPrintPeriod = period;

        this.logln("Set default print period to " + period);
    }

    /**
     * Get the next fileName based on current settings.
     */
    private String getFileName(String id) {
        return String.format(this.defaultFileName, id, this.defaultExtension);
    }

    /**
     * Reset (remove) all Simulation printers, cleaning up any multirun printers
     * and disabling all defaults. Removes all user-added DataPrinters as well.
     * 
     * NOTE that this must be called in order to properly flush the output of
     * MultiRun Printers.
     */
    public void cleanupPrinters() {
        // Cleanup any user-added Printers
        for (Printer p : this.myPrinters.keySet()) {
            if (p instanceof MultiRunPrinter)
                ((MultiRunPrinter) p).cleanup();
        }

        this.dataPrinters.clear(); // user added DataPrinters
        this.myPrinters.clear(); // user added Printers
        this.defaultCSVEnabled = false; // CSV
        if (this.defaultMRP != null) { // MRP
            this.defaultMRP.cleanup();
            this.defaultMRP = null;
        }
        if (this.defaultCleanup != null) { // Cleanup
            this.defaultCleanup.cleanup();
            this.defaultCleanup = null;
        }

        this.logln("Cleaned all Printers");
    }

    /**
     * Adds a printer of class printerClass with fileName and period. This
     * Printer will be removed after one run of the Economy unless it is a
     * MultiRun printer.
     */
    public void addPrinter(Class<? extends Printer> printerClass,
            String fileName, int period) {
        this.myPrinters.put(Simulations.create(printerClass, fileName, period),
                false);

        this.logln("Added Printer: class=%s, filename=%s, period=%d",
                printerClass.getSimpleName(), fileName, period);
    }

    /**
     * Enables the default CSVPrinter. Multiple runs will have _Run# appended to
     * fileName so as not to overwrite the files. This printer is enabled by
     * default.
     */
    public void enableDefaultCSVPrinter(String fileName, int period) {
        this.defaultCSVEnabled = true;

        this.logln("Enabled default CSV Printer for " + fileName);
    }

    /**
     * Disables the default CSVPrinter. Enabled by default.
     */
    public void disableDefaultCSVPrinter() {
        this.defaultCSVEnabled = false;

        this.logln("Disabled default CSV Printer");
    }

    /**
     * Enables the default CSVMultiRunPrinter with fileName and period.
     */
    public void enableDefaultMultiRunPrinter(String fileName, int period) {
        if (this.defaultMRP == null
                || !this.defaultMRP.getFileName().equals(fileName)
                || this.defaultMRP.getPrintPeriod() != period) {

            if (this.defaultMRP != null)
                this.defaultMRP.cleanup();
            this.defaultMRP = new CSVMultiRunPrinter(fileName, period);
            this.defaultMRP.addDataPrinters(this.getCSVMultiRunDataPrinters());
        }

        this.logln("Enabled default MultiRun Printer for " + fileName);
    }

    /**
     * Disables the default CSVMultiRunPrinter, performing any necessary
     * cleanup. Disabled by default.
     */
    public void disableDefaultMultiRunPrinter() {
        if (this.defaultMRP != null) {
            this.defaultMRP.cleanup();
            this.defaultMRP = null;

            this.logln("Disabled default MultiRun Printer");
        } else
            this.logln("Disabled default MultiRun Printer (already disabled)");
    }

    /**
     * Enables the default MultiRunCleanupPrinter for fileName. Should be called
     * after adding Agents and Markets.
     */
    public void enableDefaultCleanupPrinter(String fileName) {
        if (this.defaultCleanup == null
                || !this.defaultCleanup.getFileName().equals(fileName)) {

            if (this.defaultCleanup != null)
                this.defaultCleanup.cleanup();
            this.defaultCleanup = new CSVMultiRunCleanupPrinter(fileName);
            this.defaultCleanup.addDataPrinters(this.getCleanupDataPrinters());
        }

        this.logln("Enabled default Cleanup Printer for " + fileName);
    }

    /**
     * Disable the default MultiRunCleanupPrinter, performing any necessary
     * cleanup. Disabled by default.
     */
    public void disableDefaultCleanupPrinter() {
        if (this.defaultCleanup != null) {
            this.defaultCleanup.cleanup();
            this.defaultCleanup = null;

            this.logln("Disabled default Cleanup Printer");
        } else
            this.logln("Disabled default Cleanup Printer (already disabled)");
    }

    /**
     * Adds dp to the default CSVPrinter. If dp is not multiRun compatible, it
     * will be removed after one run.
     */
    public void addDataPrinter(DataPrinter dp) {
        this.dataPrinters.put(dp, null);

        this.logln("Added DataPrinter %s to the default CSV Printer", dp
                .getClass().getSimpleName());
    }

    /**
     * Adds dp to the Printer printing to fileName (as its base fileName). If dp
     * is not multiRun compatible, it will be removed after one run.
     */
    public void addDataPrinter(DataPrinter dp, String fileName) {
        this.dataPrinters.put(dp, fileName);

        this.logln("Added DataPrinter %s for %s",
                dp.getClass().getSimpleName(), fileName);

    }

    /**
     * @param economy
     *            the initial economy
     * @return a set of basic multi-run, CSV compatible DataPrinters for this
     *         simulation
     */
    public Collection<DataPrinter> getCSVMultiRunDataPrinters() {
        Collection<DataPrinter> dps = new ArrayList<DataPrinter>();

        dps.add(new TimeStepDataPrinter()); // time step
        dps.add(new NumberAliveDataPrinter()); // population

        // market prices
        for (MarketInfo mi : this.marketInfos)
            dps
                    .add(new MarketPriceDataPrinter(mi.currencyName,
                            mi.productName));

        // market volumes
        for (MarketInfo mi : this.marketInfos)
            dps
                    .add(new MarketVolumeDataPrinter(mi.currencyName,
                            mi.productName));

        return dps;
    }

    /**
     * @param economy
     *            the initial economy
     * @return a set of basic CSV compatible DataPrinters for this simulation
     */
    public Collection<DataPrinter> getCSVDataPrinters() {
        Collection<DataPrinter> dps = this.getCSVMultiRunDataPrinters();

        // Good Trackers
        // 1) Get the Set of classes used in this Simulation sorted by name
        SortedMap<String, Class<? extends Agent>> classes = new TreeMap<String, Class<? extends Agent>>();
        for (AgentInfo ai : this.agentInfos)
            classes.put(ai.cls.getSimpleName(), ai.cls);

        // 2) For each Class/Good make an avg GoodTracker dp
        for (Class<? extends Agent> cls : classes.values())
            for (String goodName : this.goods.keySet()) {
                dps.add(new GoodTrackerDataPrinter(goodName, cls, true));
            }

        return dps;
    }

    /**
     * @param economy
     *            the initial economy
     * @return a set of basic CSV, multi-run, cleanup compatible DataPrinters
     *         for this simulation
     */
    public Collection<DataPrinter> getCleanupDataPrinters() {
        Collection<DataPrinter> dps = this.getCSVMultiRunDataPrinters();
        DataPrinter timeStep = null; // remove the time step DataPrinter
        for (DataPrinter dp : dps)
            if (dp instanceof TimeStepDataPrinter) {
                timeStep = dp;
                break;
            }
        Errors.Assert(timeStep == null || dps.remove(timeStep));

        return dps;
    }
}