/**
 * 
 */
package eos.markets;

import java.util.Collections;
import java.util.Iterator;

import eos.economies.BL321Economy;
import eos.framework.*;
import eos.models.NameAndQuantityGood;
import eos.utilities.Errors;

/**
 * The SellerPostedPriceMarket is a simple, retail type Market where buyers
 * choose to take or leave Goods based on seller prices. When offers are added
 * to the Market, it sets aside the offered currency and product until the
 * Market is cleared. This prevents Agents from unduly influencing Market
 * outcomes by offering bids that they cannot honor.
 * 
 * @author Michael Adelson
 */
public class SellerPostedPriceMarket extends CallAuctionMarket {

    /**
     * Construct a new Seller's Market.
     */
    public SellerPostedPriceMarket(Economy economy, String currencyName,
            String productName) {
        super(economy, currencyName, productName);
    }

    /**
     * Clears buyers against seller posted prices. Buyer bids are shuffled, not
     * sorted, before beginning.
     */
    public void clear() {
        // reset market volume
        this.marketVolume = 0.0;

        // empty offer list -> no transactions
        if (buyOffers.isEmpty() || sellOffers.isEmpty()) {
            // return products to failed sellers
            for (Offer o : sellOffers)
                o.product.deposit(o.maxQuantity);
            // return currency to failed buyers
            for (Offer o : buyOffers)
                o.currency.deposit(o.maxQuantity * o.price);

            // reset offer lists
            buyOffers.clear();
            sellOffers.clear();
            return;
        }

        // shuffle and sort
        Collections.shuffle(this.buyOffers, this.economy.getRandom());
        Collections.shuffle(this.sellOffers, this.economy.getRandom());
        Collections.sort(this.sellOffers);

        /*
         * For each buyer, buy as much as possible from the list of sellers.
         */
        double weightedAvgPrice = 0.0;
        Iterator<Offer> buyers = this.buyOffers.iterator(), sellers = this.sellOffers
                .iterator();
        Offer buy = buyers.next(), sell = sellers.next(); // already checked for
        // empty lists above
        do {
            // if the buy price is too low, or there are no sell offers, resolve
            // the buy offer and move on.
            if (sell == null || less(buy.price, sell.price)) {
                buy.currency.deposit(buy.maxQuantity * buy.price);
                buy = buyers.hasNext() ? buyers.next() : null;
            } else { // buy at sell price
                double priorVol = this.marketVolume; // incremented by transact
                boolean sellExhausted = this.transact(sell, buy, sell.price);

                // maintain marketPrice weighted by volume
                double deltaVol = this.marketVolume - priorVol;
                weightedAvgPrice += sell.price * deltaVol;

                // advance the offers
                if (sellExhausted)
                    sell = sellers.hasNext() ? sellers.next() : null;
                else
                    buy = buyers.hasNext() ? buyers.next() : null;
            }
        } while (buy != null); // when this exits, all buy offers are finished

        // return unsold product
        while (sell != null) {
            sell.product.deposit(sell.maxQuantity);
            sell = sellers.hasNext() ? sellers.next() : null;
        }

        // finalize marketPrice - don't update if marketVolume = 0
        if (less(0.0, this.marketVolume))
            this.marketPrice = weightedAvgPrice / this.marketVolume;

        // clear offer books
        this.buyOffers.clear();
        this.sellOffers.clear();
    }

    /*---------------------- TEST CLIENT ------------------------*/

    /**
     * Test client
     */
    public static void main(String[] args) {
        final int numAgents = 20;
        final int sq = 10;

        NameAndQuantityGood.registerName(NameAndQuantityGood.class, "Food");
        NameAndQuantityGood.registerName(NameAndQuantityGood.class, "Money");

        Good[] products = new Good[numAgents];
        Good[] currencies = new Good[numAgents];
        for (int i = 0; i < numAgents; i++) {
            products[i] = new NameAndQuantityGood("Food");
            products[i].deposit(sq);
            currencies[i] = new NameAndQuantityGood("Money");
            currencies[i].deposit(sq);
        }

        BL321Economy e = new BL321Economy();
        e.getRandom().setSeed(123);
        SellerPostedPriceMarket m = new SellerPostedPriceMarket(e, "Money",
                "Food");

        // same price sell
        for (int i = 0; i < numAgents / 2; i++) {
            m.addSellOffer(currencies[i], products[i], 1, 1);
        }
        for (int i = numAgents / 2; i < numAgents; i++) {
            m.addBuyOffer(currencies[i], products[i], 1, 1);
        }
        m.clear();

        Errors.Assert(m.equal(m.getLastMarketPrice(), 1));
        Errors.Assert(m.equal(m.getLastMarketVolume(), numAgents / 2));

        for (int i = 0; i < numAgents / 2; i++) {
            Errors.Assert(m.equal(currencies[i].getQuantity(), sq + 1));
            Errors.Assert(m.equal(products[i].getQuantity(), sq - 1));
            Errors.Assert(m.equal(products[i + numAgents / 2].getQuantity(),
                    sq + 1));
            Errors.Assert(m.equal(currencies[i + numAgents / 2].getQuantity(),
                    sq - 1));
        }

        System.out.println("Single price test complete");

    }
}
