/**
 * 
 */
package eos.models;

import java.util.*;

import eos.framework.*;
import eos.goods.EmptyGood;
import eos.utilities.Errors;

/**
 * A class defining the BaseLine3 Model. Contains the model level Agents and
 * Goods for BaseLine3.
 * 
 * This is a new way of defining an EOS model. Although it makes for a lengthy
 * java file, all related classes are grouped in one place. Alternatively, these
 * classes could be placed in a package of their own. As models get larger, that
 * method will probably become preferable. For now, though, this method provides
 * a simple way to encapsulate the essence of a model.
 * 
 * @author Michael Adelson
 */
public class Baseline3Model {

    /*------------------------------ Constants ------------------------------*/

    /**
     * Arbitrary values representing prices before the simulation began. Agents
     * may wish to use these values as a basis for initial bidding.
     */
    public static double HISTORICAL_FOOD_PRICE = 0.2,
            HISTORICAL_LABOR_PRICE = 0.4, HISTORICAL_UTILITY_PRICE = 1.0;

    /**
     * The value of leisure in terms of utility
     */
    public static final double LEISURE_VALUE = 5.0;

    /*----------------- BaseLine2 Goods ----------------------*/

    /**
     * The Goods used in BL3
     * 
     * @author Michael Adelson
     */
    public static enum Goods {
        Food, Labor, Money, Utility;
    }

    /**
     * Defines a fixed set of Goods for use in the BaseLine2 economy.
     * 
     * @author Michael Adelson
     */
    public static class Baseline3Good extends NameAndQuantityGood {

        /**
         * Register allowed Goods with superclass.
         */
        static {
            for (Goods goodName : Goods.values())
                NameAndQuantityGood.registerName(Baseline3Good.class, goodName
                        .toString());
        }

        /**
         * Construct a BaseLine2Good with name.
         */
        public Baseline3Good(String name) {
            super(name);
        }
    }

    /*----------------- BaseLine2 Agents ----------------------*/

    /**
     * Abstract class which sets up the Goods and Markets used in the Baseline3
     * economy.
     * 
     * @author Mike Adelson
     */
    public static abstract class Baseline3Laborer extends OwnerLaborer {

        // This Agent's goods
        protected final Good food, labor, money, utility;

        // This Agent's markets
        protected final Market foodMarket, laborMarket, utilityMarket;

        /**
         * Construct a BaseLineAgent.
         */
        public Baseline3Laborer(Economy economy, Map<String, Good> goods) {
            super(economy, goods);

            // Goods
            this.food = goods.get(Goods.Food.toString());
            this.labor = goods.get(Goods.Labor.toString());
            this.money = goods.get(Goods.Money.toString());
            this.utility = goods.get(Goods.Utility.toString());

            Errors.AssertHasGoods(this, EnumSet.allOf(Goods.class));

            // Markets
            this.foodMarket = this.economy.getMarketFor(this.money.getName(),
                    this.food.getName());
            this.laborMarket = this.economy.getMarketFor(this.money.getName(),
                    this.labor.getName());
            this.utilityMarket = this.economy.getMarketFor(
                    this.money.getName(), this.utility.getName());
            Errors.Assert(this.foodMarket != null && this.laborMarket != null
                    && this.utilityMarket != null);
        }

        /**
         * Calls baseLine2LaborerSubclassAct(), THEN converts any unsold labor
         * to leisure. Note that ANY labor which the laborer even tries to sell
         * will not be converted to leisure, assuming that the Market used sets
         * aside all bid labor.
         */
        protected final void ownedLaborerSubclassAct() {
            this.utility.deposit(LEISURE_VALUE
                    * (Laborer.LABOR_ALLOWANCE - this.getLastLaborSold()));

            this.baseLine3LaborerSubclassAct();
        }

        /**
         * Should be implemented by subclasses to perform further actions when
         * act() is called. Called BEFORE labor is converted to leisure.
         */
        abstract protected void baseLine3LaborerSubclassAct();

        /**
         * Returns a List of food, labor, money, and leisure.
         */
        public Collection<Good> getGoods() {
            Collection<Good> goods = new ArrayList<Good>();
            goods.add(this.food);
            goods.add(this.labor);
            goods.add(this.money);
            goods.add(this.utility);

            return goods;
        }

        /**
         * Returns food, labor, money, or leisure. Or, throws an exception.
         */
        public Good getGood(String goodName) {
            switch (Goods.valueOf(goodName)) {
            case Money:
                return this.money;
            case Food:
                return this.food;
            case Labor:
                return this.labor;
            case Utility:
                return this.utility;
            default:
                throw new RuntimeException("Agent does not have Good "
                        + goodName);
            }
        }

        /**
         * Returns the Agent's quantity of money.
         */
        public double getNetWorth() {
            return this.money.getQuantity();
        }

        /**
         * Get this Agent's food (required by Laborer)
         */
        protected Good getFood() {
            return this.food;
        }

        /**
         * Get this Agent's labor (required by Laborer)
         */
        protected Good getLabor() {
            return this.labor;
        }
    }

    /**
     * Abstract class which sets up the Goods and Markets used in the Baseline3
     * 
     * @author Michael Adelson
     */
    public static abstract class Baseline3Firm extends OwnedFirm {

        // This Agent's goods
        protected final Good labor, money, product;

        // This Agent's markets
        protected final Market laborMarket, productMarket;

        /**
         * True if this Firm represents a food-producing farm.
         */
        protected final boolean isFarm;

        /**
         * Construct a BaselineFirm
         */
        public Baseline3Firm(Economy economy, Map<String, Good> goods) {
            super(economy, goods);

            // Goods
            this.labor = goods.get(Goods.Labor.toString());
            this.money = goods.get(Goods.Money.toString());
            Good food = goods.get(Goods.Food.toString());
            Good utility = goods.get(Goods.Utility.toString());
            if (food.getQuantity() > 0.0 && utility.getQuantity() <= 0.0)
                this.product = food; // a farm
            else if (utility.getQuantity() > 0.0)
                this.product = utility; // a luxury factory
            else {
                Errors.warn("BL3 Firm has " + food.getQuantity()
                        + " starting food, " + utility.getQuantity()
                        + " starting utility. It is being made a farm");
                this.product = food;
            }

            this.isFarm = this.product.getName().equals(Goods.Food.toString());
            Errors.AssertHasGoods(this, EnumSet.of(Goods.Money, Goods.Labor,
                    Goods.valueOf(this.product.getName())));

            // Markets
            this.laborMarket = this.economy.getMarketFor(this.money.getName(),
                    this.labor.getName());
            this.productMarket = this.economy.getMarketFor(
                    this.money.getName(), this.product.getName());
            Errors.Assert(this.productMarket != null
                    && this.laborMarket != null);
        }

        /**
         * The implementation of production. Converts available Labor to Food.
         * 
         * @return the amount produced.
         */
        protected final double produce() {
            double amountProduced;
            amountProduced = convertToProduct(this.labor.withdraw(this.labor
                    .getQuantity()));
            this.product.deposit(amountProduced);

            return amountProduced;
        }

        /**
         * @return the amount of product produced from laborQuantity.
         */
        protected double convertToProduct(double laborQuantity) {
            if (this.isFarm)
                return convertToFood(laborQuantity);
            return convertToUtility(laborQuantity);
        }

        /**
         * This Firm is only Bankrupt when it has no labor, money, or product
         * left. Note that this call is only necessarily valid at the beginning
         * or end of a time step.
         */
        public boolean isBankrupt() {
            return this.money.getQuantity() <= 0.0
                    && this.labor.getQuantity() <= 0.0
                    && this.product.getQuantity() <= 0.0;
        }

        /**
         * Calls baseLine3FirmSubclassAct().
         * 
         * @see OwnedFirm#ownedFirmSubclassAct()
         */
        protected final void ownedFirmSubclassAct() {
            this.baseLine3FirmSubclassAct();
        }

        /**
         * Should be implemented by subclasses to perform further actions during
         * act().
         * 
         * @see Baseline3Firm#ownedFirmSubclassAct()
         */
        abstract protected void baseLine3FirmSubclassAct();

        /**
         * Returns a List of food, labor, money.
         */
        public Collection<Good> getGoods() {
            Collection<Good> goods = new ArrayList<Good>();
            goods.add(this.labor);
            goods.add(this.money);
            goods.add(this.product);

            return goods;
        }

        /**
         * No activation costs defined.
         */
        public final Map<String, Double> getRemainingActivationCosts() {
            Errors.err("No activation costs defined");
            return new HashMap<String, Double>();
        }

        /**
         * Returns food, labor, or money, or throws an exception.
         */
        public Good getGood(String goodName) {
            switch (Goods.valueOf(goodName)) {
            case Money:
                return this.money;
            case Food:
                if (!this.isFarm)
                    return EmptyGood.get(goodName);
                return this.product;
            case Labor:
                return this.labor;
            case Utility:
                if (this.isFarm)
                    return EmptyGood.get(goodName);
                return this.product;
            default:
                throw new RuntimeException("Agent does not have Good " + goodName);
            }
        }

        /**
         * Returns the Agent's quantity of money.
         */
        public double getNetWorth() {
            return this.money.getQuantity();
        }

        /**
         * Get this Agent's labor (required by Firm)
         */
        protected Good getLabor() {
            return this.labor;
        }

    }

    /*------------------------------ Production ------------------------------*/

    /**
     * The production function for Food in BaseLine2: MIN(laborQuantity *
     * LABOR_PRODUCTIVITY, MAX_PRODUCTION).
     */
    public static double convertToFood(double laborQuantity) {
        return (-0.0652 * laborQuantity * laborQuantity) + 4.0652
                * laborQuantity;
    }

    /**
     * The conversion from Labor to Utility
     */
    public static double convertToUtility(double laborQuantity) {
        return convertToFood(laborQuantity) / 2.0;
    }

    /**
     * Get the historical price for Good.
     * 
     * @see Baseline3Model#HISTORICAL_FOOD_PRICE
     */
    public static double getHistoricalPrice(Good good) {
        switch (Goods.valueOf(good.getName())) {
        case Food:
            return HISTORICAL_FOOD_PRICE;
        case Labor:
            return HISTORICAL_LABOR_PRICE;
        case Utility:
            return HISTORICAL_UTILITY_PRICE;
        default:
            throw new RuntimeException("No historical price for Good "
                    + good.getName());
        }
    }

    /*----------------- Constructor ---------------------------*/

    /**
     * Can't be instantiated.
     */
    private Baseline3Model() {
    }
}
