package eos.models;

import eos.framework.Good;

/**
 * ContinousGood provides a model-level implementation of all methods for a 
 * Good that should be stored in double quantities. Classes that extend 
 * ContinousGood need not implement any additional methods in order to be 
 * compatible with Markets.
 * 
 * Author: Michael Adelson 3-29-09
 */
public abstract class ContinuousGood implements Good {

    // The quantity represented by this Good
    private double quantity;
    
    /**
     * Construct a ContinousGood with 0 initial quantity.
     */
    public ContinuousGood() {
        this.quantity = 0;
    }
    
    /**
     * Add amount of this good to quantity. A RuntimeException will be 
     * thrown if amount is negative.
     */
    public void deposit(double amount) {
        if (amount < 0) throw new RuntimeException("Amount must be >= 0");
        quantity += amount;
    }

    /**
     * Returns the quantity of this good contained.
     */
    public double getQuantity() {
        return quantity;
    }

    /**
     * Remove amount of this good. If this ContinousGood's quantity is <
     * amount, quantity is set to 0. Returns the amount of good successfully 
     * withdrawn. A RuntimeException will be thrown if amount is negative. 
     */
    public double withdraw(double amount) {
        if (amount < 0) throw new RuntimeException("Amount must be >= 0");
        if (amount <= quantity) {
            quantity -= amount;
            return amount;
        }
        else {
            double oldQuantity = quantity;
            quantity = 0;
            return oldQuantity;
        }
    }
    
    /**
     * Returns false, since this Good's quantity is represented as a double.
     */
    public final boolean isDiscrete() { return false;}
}
