/**
 * edit 3-24-2009 Chris Rucinski
 * edit 11-04-2009 Michael Adelson
 */

package eos.models;

import java.util.Map;
import eos.framework.*;
import eos.utilities.Errors;

/**
 * A general model for a Firm. Act() calls produce(), discardLabor(),
 * checkIfBankrupt(), and finally firmSubclassAct(). This implementation makes
 * no model assumptions except for the existence of a Good representing labor,
 * which Firm accesses throught the abstract getLabor() function.
 * 
 * @author Chris Rucinski
 * @author Michael Adelson
 */
public abstract class Firm extends Agent {

    /**
     * How much did this Firm produce this time step?
     */
    private double lastAmountProduced;

    /**
     * How much labor did this Firm have to work with at the beginning of this
     * time step.
     */
    private double lastStartingLabor;

    /**
     * Create a new Firm in economy.
     */
    public Firm(Economy economy, Map<String, Good> goods) {
        super(economy, goods);
    }

    /**
     * Firms are a more specific type of agent in that they can convert labor
     * into other goods and lose any unused labor at the end of each timestep.
     */
    public final void act() {
        Errors.Assert(this.isAlive(), "Must be alive to act()");
        this.lastStartingLabor = this.getLabor().getQuantity();
        this.lastAmountProduced = this.produce();
        this.discardLabor();
        this.checkIfBankrupt();
        this.firmSubclassAct();
    }

    /**
     * Actions to be performed by the subclass of Firm.
     */
    abstract protected void firmSubclassAct();

    /**
     * Check if firm is bankrupt; the firm should die if it is in fact bankrupt.
     */
    abstract protected void checkIfBankrupt();

    /**
     * Set the amount of labor that this firm has to 0. Can be overridden if
     * necessary.
     */
    protected void discardLabor() {
        this.getLabor().withdraw(this.getLabor().getQuantity() + 1.0);
    }

    /**
     * @return How much this Firm produced during this time step.
     */
    public final double getLastAmountProduced() {
        return this.lastAmountProduced;
    }

    /**
     * @return How much labor this Firm had at the beginning of this time step.
     */
    public final double getLastStartingLabor() {
        return this.lastStartingLabor;
    }

    /**
     * Should return a Good representing the Firm's labor. Called by the default
     * implementation of discardLabor().
     */
    protected abstract Good getLabor();

    /**
     * Convert labor into another good
     * 
     * @return the amount produced
     */
    abstract protected double produce();
}
