/**
 * 
 */
package eos.simulations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Random;

import eos.agents.*;
import eos.economies.BL321Economy;
import eos.framework.*;
import eos.io.*;
import eos.io.AgentTrackerDataPrinter.AgentTrackerMode;
import eos.markets.CallAuctionMarket;
import eos.models.Baseline2Model;
import eos.models.Baseline2Model.Baseline2Good;
import eos.models.SingleOwnerFirms.*;
import eos.utilities.*;
import eos.utilities.Simulations.NameClsMap;

/**
 * @author Michael Adelson
 * 
 */
public class BL2Sim {

    // statics

    // classes

    static Class<? extends Agent> LaborerClass = BL2FixedPriceLaborer.class,
            OwnerClass = BL2LazyOwner.class,
            FarmClass = BL2FixedPriceFarm.class,
            TraderClass = SimpleTrader.class;

    static Class<? extends Market> MarketClass = CallAuctionMarket.class;

    // initial conditions
    static NameClsMap goods = Simulations.getMap("Food", "Labor", "Money",
            "Leisure", Baseline2Good.class);
    static double[] qSL = { 30, 0, 30, 0 };
    static double[] qSF = { 0, 0, 100, 0 };
    static double[] qST = { 20, 0, 100, 0 };

    static int laborersPerFarm = 60;
    static int startingFarms = 1;
    static int startingLaborers = laborersPerFarm * startingFarms;
    static int startingTraders = 0;

    static int steps = 5000;
    static int printPeriod = 50;

    /**
     * @param args
     */
    public static void main(String[] args) {
        Baseline2Model.HISTORICAL_LABOR_PRICE = (59.0 / 24.0)
                * Baseline2Model.HISTORICAL_FOOD_PRICE;
        // Baseline2Model.HISTORICAL_LABOR_PRICE = (4.0 / 5.0) *
        // Baseline2Model.HISTORICAL_LABOR_PRICE;
        Random r = new Random();
        BL321Economy e = getOwnerEconomy(r.nextLong());
        // F09Economy e = getEconomy(r.nextLong());
        perturbFoodAndMoney(e, 5, r);
        Printer p = getPrinter(printPeriod, "fpBL2.csv");
        p.addDataPrinter(new GoodTrackerDataPrinter("Money", false));
        e.addPrinter(p);
        e.run(steps);
        p.cleanup(e);
    }

    /**
     * @return the number of living Agents at the end of the simulation.
     */
    public static int run(BL321Economy economy, String fileName, int agentsToTrack) {

        if (fileName != null) {
            Printer printer = getPrinter(printPeriod, fileName);
            printer.addDataPrinters(getAgentTrackers(economy, agentsToTrack));
            economy.addPrinter(printer);
        }
        economy.run(steps);
        economy.cleanupPrinters();

        return economy.getAgents().size();
    }

    public static BL321Economy getEconomy(long seed) {
        return getEconomy(seed, startingLaborers, startingFarms,
                startingTraders);
    }

    public static BL321Economy getEconomy(long seed, int laborers, int farms,
            int traders) {
        BL321Economy E = new BL321Economy();
        E.getRandom().setSeed(seed);
        // initialize E

        // Markets
        // add a Food for Money Market
        E.addMarket(Simulations.createMarket(MarketClass, "Money", "Food", E));

        // add a Labor for Money Market
        E.addMarket(Simulations.createMarket(MarketClass, "Money", "Labor", E));

        // Agents
        ArrayList<Agent> agents = new ArrayList<Agent>(laborers + farms
                + traders);

        // add Laborers
        agents.addAll(Simulations.createAgents(LaborerClass, goods, qSL, E, laborers));

        // add Farms
        agents.addAll(Simulations.createAgents(FarmClass, goods,
                qSF, E, farms));

        // add Traders
        agents.addAll(Simulations.createAgents(TraderClass, goods,
                qST, E, traders));

        E.addAgents(agents);

        return E;
    }

    public static BL321Economy getOwnerEconomy(long seed) {
        return getOwnerEconomy(seed, startingLaborers, startingFarms);
    }

    public static BL321Economy getOwnerEconomy(long seed, int laborers, int farms) {
        BL321Economy E = new BL321Economy();
        E.getRandom().setSeed(seed);
        // initialize E

        // Markets
        // add a Food for Money Market
        E.addMarket(Simulations.createMarket(MarketClass, "Money", "Food", E));

        // add a Labor for Money Market
        E.addMarket(Simulations.createMarket(MarketClass, "Money", "Labor", E));

        // Agents
        List<? extends Agent> workers, firms, owners;

        // add Laborers
        workers = Simulations.createAgents(LaborerClass, goods,
                qSL, E, laborers - farms);

        // add Farms
        firms = Simulations.createAgents(FarmClass, goods, qSF, E,
                farms);

        // add Owners: 1 per farm
        owners = Simulations.createAgents(OwnerClass, goods, qSL,
                E, farms);

        // link owners with firms:
        for (int i = 0; i < farms; i++)
            SingleOwnerFirm.class.cast(firms.get(i)).transferOwnership(
                    SingleOwnerFirmOwner.class.cast(owners.get(i)));

        E.addAgents(workers);
        E.addAgents(firms);
        E.addAgents(owners);

        return E;
    }

    /**
     * @param period
     *            the period of the Printer
     * @param fileName
     *            the fileName to print to
     * @return a fully set up printer with period period
     */
    public static Printer getPrinter(int period, String fileName) {
        CSVPrinter printer = new CSVPrinter(fileName, period);

        printer.addDataPrinter(new TimeStepDataPrinter());
        printer.addDataPrinter(new NumberAliveDataPrinter(LaborerClass, false));
        printer.addDataPrinter(new MarketPriceDataPrinter("Food"));
        printer.addDataPrinter(new MarketPriceDataPrinter("Labor"));
        printer.addDataPrinter(new MarketVolumeDataPrinter("Food"));
        printer.addDataPrinter(new MarketVolumeDataPrinter("Labor"));

        for (String goodName : goods.keySet()) {
            if (goodName.equals("Labor"))
                continue;
            printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                    LaborerClass, true));
            printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                    LaborerClass, false));
            if (!goodName.equals("Leisure")) {
                printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                        FarmClass, true));
                printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                        FarmClass, false));
            }
            printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                    OwnerClass, true));
            printer.addDataPrinter(new GoodTrackerDataPrinter(goodName,
                    OwnerClass, false));
        }

        return printer;
    }

    /**
     * @param economy
     *            the economy to use
     * @param laborers
     *            the number of laborers to track
     * @param farms
     *            the number of farms to track
     * @param traders
     *            the number of traders to track
     * @return AgentTrackers for the first laborers laborers and farms farms
     */
    public static Collection<DataPrinter> getAgentTrackers(
            DataOutputEconomy economy, int laborers, int farms, int traders) {
        Collection<DataPrinter> dps = new ArrayList<DataPrinter>();

        // farms
        for (int i = 0; i < farms; i++) {
            dps.add(new AgentTrackerDataPrinter(FarmClass, i, economy,
                    AgentTrackerMode.Good, "Food"));
            dps.add(new AgentTrackerDataPrinter(FarmClass, i, economy,
                    AgentTrackerMode.Good, "Money"));
        }

        // laborers
        for (int j = 0; j < laborers; j++) {
            dps.add(new AgentTrackerDataPrinter(LaborerClass, j, economy,
                    AgentTrackerMode.Good, "Food"));
            dps.add(new AgentTrackerDataPrinter(LaborerClass, j, economy,
                    AgentTrackerMode.Good, "Money"));
        }

        // laborers
        for (int j = 0; j < traders; j++) {
            dps.add(new AgentTrackerDataPrinter(TraderClass, j, economy,
                    AgentTrackerMode.Good, "Food"));
            dps.add(new AgentTrackerDataPrinter(TraderClass, j, economy,
                    AgentTrackerMode.Good, "Money"));
        }

        return dps;
    }

    public static Collection<DataPrinter> getAgentTrackers(
            DataOutputEconomy economy, int numAgents) {
        Collection<DataPrinter> dps = new ArrayList<DataPrinter>();

        for (int j = 0; j < numAgents; j++) {
            dps.add(new AgentTrackerDataPrinter(j, economy,
                    AgentTrackerMode.Good, "Food"));
            dps.add(new AgentTrackerDataPrinter(j, economy,
                    AgentTrackerMode.Good, "Money"));
        }

        return dps;
    }

    /**
     * @return a String representation of the initial conditions.
     */
    public static String getInitialConditions() {
        Object[] agoods = goods.keySet().toArray();
        String s = "\t\t", n = "\n";
        String out = "Initial Conditions:" + n;
        out += "Class" + s + "Default Pop" + s + agoods[0] + s + agoods[1] + s
                + agoods[2] + n;
        out += Strings.getShortNames(LaborerClass) + s + +startingLaborers + s
                + qSL[0] + s + qSL[1] + s + qSL[2] + n;
        out += Strings.getShortNames(FarmClass) + s + startingFarms + s + qSF[0]
                + s + qSF[1] + s + qSF[2] + n;
        if (startingTraders > 0)
            out += Strings.getShortNames(TraderClass) + s + startingTraders + s
                    + startingFarms + s + qSF[0] + s + qSF[1] + s + qSF[2] + n;

        return out;
    }

    public static void perturbFoodAndMoney(BL321Economy economy,
            int maxPerturbation, Random r) {
        for (Agent a : economy.getAgents()) {
            double f = r.nextDouble() * maxPerturbation;
            if (r.nextBoolean())
                a.getGood("Food").withdraw(f);
            else
                a.getGood("Food").deposit(f);

            double m = r.nextDouble() * maxPerturbation;
            if (r.nextBoolean())
                a.getGood("Money").withdraw(m);
            else
                a.getGood("Money").deposit(m);
        }
    }

}
